// Author: Pete Kirkham
//
// quad store utils.
//
// string table.
// simpler than a general hash map, as it can only grow.
//
#include <iostream>

#include "utils/string_table.h"

class string_table_entry
{
public:
    string_table_entry (unsigned int hash, const std::string value) : _hash(hash), _value(value) { }
    unsigned int    _hash;
    std::string     _value;
};

string_table::string_table    () : _entries(NULL), _size(0), _allocated(0), _mask(0)
{
    _recent = (string_table_entry**) malloc(8 * sizeof(string_table_entry*));

    for (size_t i(0); i < 8; ++i)
    {
        _recent[i] = NULL;
    }

    resize(16 * KiB);
}

string_table::~string_table   ()
{
    for (size_t i(0); i < _allocated; ++i)
    {
        delete _entries[i];
    }

    free(_entries);
    free(_recent);
}

bool    string_table::load    (const std::string& path)
{
    return false;
}

bool    string_table::save    (const std::string& path)
{
    return false;
}

unsigned int    get_hash              (const std::string& string)
{
    unsigned int hash(0);

    for (size_t i(0), l(string.length()); i < l; ++i)
    {
        hash = hash * 17 + string[i];
    }

    return mix(hash);
}

unsigned int    string_table::put     (const std::string& string)
{
    if (_size * 8 > _allocated)
    {
        resize(_allocated * 2);
        std::cerr << "strings: " << _size << std::endl;
    }

    unsigned int    hash(get_hash(string));
    unsigned int    mask(_mask);
    bool            checked_recent(false);

    for (;;)
    {
        size_t              index(hash & mask);
        string_table_entry* entry(_entries[index]);

        if (entry)
        {
            if ((hash == entry->_hash) && (string == entry->_value))
            {
                _recent[hash&0x7] = entry;
                return hash;
            }

            if (!checked_recent)
            {
                // check recent
                for (size_t index(0); index < 8; ++index)
                {
                    string_table_entry* entry(_recent[index]);

                    if ((entry) && (hash == entry->_hash) && (string == entry->_value))
                    {
                        std::cerr << "from recent" << std::endl;
                        return hash;
                    }
                }

                checked_recent = true;
            }
        }
        else
        {
            _recent[hash&0x7] = _entries[index] = new string_table_entry(hash, string);
            ++_size;
            return hash;
        }

        ++hash;
    }

    return hash;
}

const char*   string_table::get     (unsigned int hash)
{
    size_t              index(hash & _mask);
    string_table_entry* entry(_entries[index]);

    return entry ? entry->_value.c_str() : NULL;
}

void    string_table::resize  (size_t allocate)
{
    string_table_entry**    old_entries(_entries);

    _entries = (string_table_entry**) malloc(allocate * sizeof(string_table_entry*));

    for (size_t i(0); i < _allocated; ++i)
    {
        _entries[i] = old_entries[i];
    }

    for (size_t i(_allocated); i < allocate; ++i)
    {
        _entries[i] = NULL;
    }

    _allocated = allocate;
    _mask = allocate - 1;
}

void    string_table::dump ()
{
    for (size_t i(0); i < _allocated; ++i)
    {
        string_table_entry* entry(_entries[i]);

        if (entry)
        {
            std::cout << entry->_hash << " => \"" << entry->_value << '"' << std::endl;
        }
    }
}
